/*
 * aln_read.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Alignment Mismatch Finder source code,
 * and may only be used, modified, and distributed under the terms of
 * the license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C



/*
 * Reads axt-formatted alignment from 'file' into 'aln'.
 * Returns non-0 if end of file is reached, 0 otherwise.
 */
int load_axt_alignment(taln *aln,tfile *file)
{
    char *c,*d;

    assert(aln);
    assert(file);

    skip_line_breaks(file);
    str_add_from_file_until_line_break(&aln->info,file);
    store_char_in_str(&aln->info,0);

    skip_line_breaks(file);
    str_add_from_file_until_line_break(&aln->seq[0].data,file);

    skip_line_breaks(file);
    str_add_from_file_until_line_break(&aln->seq[1].data,file);

    c = aln->info.start;
    while (*c && !isspace(*(unsigned char*)c)) { c++; }
    while (*c && isspace(*(unsigned char*)c)) { c++; }
    d = c;
    while (*c && !isspace(*(unsigned char*)c)) { c++; }
    store_data_in_str(&aln->seq[0].name,d,c-d);
    while (*c && isspace(*(unsigned char*)c)) { c++; }
    while (*c && !isspace(*(unsigned char*)c)) { c++; }
    while (*c && isspace(*(unsigned char*)c)) { c++; }
    while (*c && !isspace(*(unsigned char*)c)) { c++; }
    while (*c && isspace(*(unsigned char*)c)) { c++; }
    d = c;
    while (*c && !isspace(*(unsigned char*)c)) { c++; }
    store_data_in_str(&aln->seq[1].name,d,c-d);

    return file_eof(file);
}

/*
 * Loads one sequence from a FASTA stream.
 * Uses buf, begins by checking for any data there.
 */
void load_fasta_sequence(tfile *file,tseq *seq)
{
    assert(file);
    assert(seq);
    assert(!seq->data.length && (seq->data.end == seq->data.start) && !seq->name.length && (seq->name.end == seq->name.start));

    skip_line_breaks(file);

    /* Checking that '>' is found. */
    if (!read_file(file)) { return; }
    if (*file->buf_start != '>') { return; }
    file->buf_start++;

    /* Reading name. */
    str_add_from_file_until_line_break(&seq->name,file);
    store_char_in_str(&seq->name,0);
    skip_line_breaks(file);

    /* Loading sequence. */
    while (read_file(file) && *file->buf_start != '>')
    {
        str_add_from_file_until_space(&seq->data,file);
        skip_spaces(file);
    }
}

/*
 * Reads the next two sequences from FASTA file and stores them in aln.
 * Returns non-0 if end of file is reached, 0 otherwise.
 */
int load_fasta_alignment(taln *aln,tfile *file)
{
    assert(aln);
    assert(file);

    load_fasta_sequence(file,&aln->seq[0]);
    load_fasta_sequence(file,&aln->seq[1]);
    return file_eof(file);
}

/*
 * Reads NSS format (name-sequence-sequence).
 * Each alignment is stored in one line, in 3 tab-separated fields:
 * alignment id, sequence 1, sequence 2.
 * Returns non-0 if end of file is reached, 0 otherwise.
 */
int load_nss_alignment(taln *aln,tfile *file)
{
    assert(aln);
    assert(file);

    skip_line_breaks(file);
    str_add_from_file_until_space(&aln->info,file);
    skip_spaces(file);
    str_add_from_file_until_space(&aln->seq[0].data,file);
    skip_spaces(file);
    str_add_from_file_until_space(&aln->seq[1].data,file);

    return file_eof(file);
}

/*
 * Reads G-Compass alignment format.
 * Returns non-0 if end of file is reached, 0 otherwise.
 */
int load_g_compass_alignment(taln *aln,tfile *file)
{
    int len;

    assert(aln);
    assert(file);

    store_data_in_str(&aln->seq[0].name,"ref",3);

    skip_line_breaks(file);

    /* Target genome. */
    str_add_from_file_until_tab_or_eol(&aln->info,file);
    store_data_in_str(&aln->seq[1].name,aln->info.start,aln->info.length);

    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Alignment ID. */

    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    len = aln->info.length;
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Reference chromosome. */
    store_data_in_str(&aln->seq[0].name,"-chr",4);
    store_data_in_str(&aln->seq[0].name,aln->info.start+len,aln->info.length-len);

    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    len = aln->info.length;
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Target chromosome. */
    store_data_in_str(&aln->seq[1].name,"-chr",4);
    store_data_in_str(&aln->seq[1].name,aln->info.start+len,aln->info.length-len);

    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    len = aln->info.length;
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Target strand. */
    store_data_in_str(&aln->seq[1].name,"_",1);
    store_data_in_str(&aln->seq[1].name,aln->info.start+len,aln->info.length-len);

    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Unused. */
    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Original alignment ID. */
    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Reference start coordinate. */
    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Reference end coordinate. */
    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Target start coordinate. */
    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Target end coordinate. */
    if (read_char_from_file(file)!='\t') { return 0; }

    str_add_from_file_until_tab_or_eol(&aln->seq[0].data,file); /* Aligned reference sequence. */
    if (read_char_from_file(file)!='\t') { return 0; }

    str_add_from_file_until_tab_or_eol(&aln->seq[1].data,file); /* Aligned reference sequence. */
    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Blastz score. */
    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Number of identical sites. */
    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Alignment length. */
    if (read_char_from_file(file)!='\t') { return 0; }

    store_char_in_str(&aln->info,' ');
    str_add_from_file_until_tab_or_eol(&aln->info,file); /* Identity (number of identical sites divided by alignment length). */

    skip_until_line_break(file);
    skip_line_breaks(file);

    store_char_in_str(&aln->info,0);
    /*fprintf(stderr,"%s\n",aln->info.start);*/

    return file_eof(file);
}

/*
 * Loads one pairwise alignment from stream F.
 * Returns: non-0 if alignment was loaded, 0 otherwise.
 */
int load_alignment(taln *aln,tfile *file)
{
    int eof = 0;
    int (*loader)(taln *,tfile *) = 0;

    assert(aln);
    assert(file);

    reset_aln(aln);

    assert(aln->seq[0].data.length==0);
    assert(aln->seq[1].data.length==0);
    assert(aln->seq[0].name.length==0);
    assert(aln->seq[1].name.length==0);
    assert(aln->seq[0].data.end == aln->seq[0].data.start);
    assert(aln->seq[0].name.end == aln->seq[0].name.start);
    assert(aln->seq[1].data.end == aln->seq[1].data.start);
    assert(aln->seq[1].name.end == aln->seq[1].name.start);

    if (file_eof(file)) { return 0; }

    switch (input_format)
    {
        case al_format_axt:       loader = &load_axt_alignment;       break;
        case al_format_fasta:     loader = &load_fasta_alignment;     break;
        case al_format_g_compass: loader = &load_g_compass_alignment; break;
        case al_format_nss:       loader = &load_nss_alignment;       break;
        default: assert(0);
    }
    eof = loader(aln,file);
    return aln->seq[0].data.length || aln->seq[1].data.length || !eof;
}

#endif
