/*
 * data.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Alignment Mismatch Finder source code,
 * and may only be used, modified, and distributed under the terms of
 * the license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C

/*
 * A simple string type.
 */
typedef struct tstr
{
    char *start;    /* Start of the string. Note that the string is NOT 0-terminated. */
    char *end;      /* Points to the character after the last one. */
    size_t size;    /* Size of currently allocated memory. */
    size_t length;  /* Length of currently stored data (equals end-start). */
    size_t alloc_step;  /* Allocation increment. */
} tstr;

/*
 * A sequence - has name and data.
 */
typedef struct tseq
{
    tstr data,name;
} tseq;

/*
 * A pairwise alignment - has two sequences and optional info line.
 */
typedef struct taln
{
    tseq seq[2];
    tstr info;
} taln;

void reset_str(tstr *str)
{
    assert(str);
    str->end = str->start;
    str->length = 0;
}

void clean_str(tstr *str)
{
    assert(str);
    if (str->start) { free1(str->start,str->size); }
    str->start = 0;
    str->end = 0;
    str->size = 0;
    str->length = 0;
}

void init_str(tstr *str,size_t alloc_step)
{
    assert(str);
    assert(alloc_step > 0);
    str->length = 0;
    str->alloc_step = alloc_step;
    str->size = alloc_step;
    str->start = malloc1(alloc_step);
    str->end = str->start;
}

void store_char_in_str(tstr* dst,char c)
{
    assert(dst);
    assert(dst->alloc_step > 0);
    if (dst->length + 1 > dst->size)
    {
        dst->start = realloc1(dst->start, dst->size + dst->alloc_step, dst->size);
        dst->end = dst->start + dst->length;
        dst->size += dst->alloc_step;
    }
    *dst->end = c;
    dst->end++;
    dst->length++;
}

void store_data_in_str(tstr* dst,char *src,size_t nbytes)
{
    size_t new_length;

    assert(dst);
    assert(dst->alloc_step > 0);
    assert(src);

    new_length = dst->length + nbytes;
    if (dst->size < new_length)
    {
        size_t new_size = dst->size;
        while (new_size < new_length) { new_size += dst->alloc_step; }
        dst->start = realloc1(dst->start,new_size,dst->size);
        dst->end = dst->start + dst->length;
        dst->size = new_size;
    }

    memcpy(dst->end,src,nbytes);
    dst->end += nbytes;
    dst->length += nbytes;
}

void reset_seq(tseq *seq)
{
    assert(seq);
    reset_str(&seq->name);
    reset_str(&seq->data);
}

void clean_seq(tseq *seq)
{
    assert(seq);
    clean_str(&seq->name);
    clean_str(&seq->data);
}

void init_seq(tseq *seq)
{
    assert(seq);
    init_str(&seq->name,NAME_ALLOC_STEP);
    init_str(&seq->data,DATA_ALLOC_STEP);
}

void reset_aln(taln *aln)
{
    assert(aln);
    reset_seq(&aln->seq[0]);
    reset_seq(&aln->seq[1]);
    reset_str(&aln->info);
}

void clean_aln(taln *aln)
{
    assert(aln);
    clean_seq(&aln->seq[0]);
    clean_seq(&aln->seq[1]);
    clean_str(&aln->info);
}

void init_aln(taln *aln)
{
    assert(aln);
    init_seq(&aln->seq[0]);
    init_seq(&aln->seq[1]);
    init_str(&aln->info,NAME_ALLOC_STEP);
}

#endif
