/*
 * globals.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Alignment Mismatch Finder source code,
 * and may only be used, modified, and distributed under the terms of
 * the license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C



/*
 * Alignment formats understood by this program.
 * Note: Some of these formats are only supported with input data, some only with output.
 */
enum al_format { al_format_unspecified,
                 al_format_axt, al_format_fasta, al_format_g_compass, al_format_nss, al_format_cmeh,
                 al_format_unknown, n_al_formats = al_format_unknown };
static const char* const al_format_names[] = {"unspecified","axt","FASTA","G-compass","nss","cmeh","unknown"};

/*
 * Input format, set from the command line (with '--input-format').
 */
static char *input_format_str = 0;
static enum al_format input_format = al_format_unspecified;

/*
 * Output format, set from the command line (with --output-format').
 */
static enum al_format output_format = al_format_unspecified;
static char *output_format_str = 0;

/*
 * Command line interface: --help and --version flags.
 */
static int opt_help = 0;
static int opt_version = 0;

/*
 * Add this number of extra alignment basepairs to each region before printing.
 * Set from the command line, with '--extend' option.
 */
static long long extend_output_regions = 0ll;

/*
 * Only scan alignments that are at least this long.
 * Controlled by '--min-aln-length' option.
 */
static long long min_alignment_length = 1ll;

/*
 * Trim both ends of all input alignments by this number of bp before processing.
 * Controlled by '--trim' option.
 */
static long long trim_alignment = 0ll;

/*
 * Set with '--use-gaps' option.
 * 0 = Gaps are ignored.
 * 1 (--use-gaps is used) = Gaps are used together with mismatches.
 */
static int use_gaps = 0;

/*
 * Mismatch-rich regions are (optionally) extended before being printed out.
 * Such extended regions may overlap.
 * When '--join' option is used, join_overlapping_regions is set to 1,
 * which means that all overlapping regions should be merged together.
 */
static int join_overlapping_regions = 0;

/*
 * If true, then any mismatches on CpG sites, such as 
 *  TACGAG
 *  TATGAG
 *    ^
 * Are ignored and not counted for variant clusters.
 */
static int ignore_CpG = 0;

/*
 * Set by '--palindrome' option.
 * Enables scanning for clusters containing just two gap clusters (one on each sequence),
 * and perfectly palindromic sequence inbetween.
 */
static int scan_for_palindromes = 0;

/*
 * Assumed evolutinary distance between sequences (mismatches / bp).
 * Set from command line with '--distance' option.
 */
static char *dist_str = 0;
static double dist = -1.0;

/*
 * Index of the first input file name in the command line.
 */
static int argv_index_of_first_path = -1;

/*
 * Scan for clusters containing min_vars to max_vars mismatches.
 * Note that the final mismatch-rich region is formed by merging overlapping or nearby clusters,
 * and may contain much more mismatches, than max_vars.
 */
static int min_vars_per_cluster = -1;
static int max_vars_per_cluster = -1;

/* E-value for detecting gap/mismatch rich region. */
static char* evalue_str = 0;
static double evalue = 0.01;




/*
 * Internals.
 */

/* How many bytes are allocated currently. */
static long long allocated = 0ll;

/* Maximum amount of memory that was allocated in current program run. */
static long long max_allocated = 0ll;




/*
 * Matrix of P(N,L).
 * P(N,L) is a probability that a random fragment of the alignment with length L contains at least
 * N mismatches or gap blocks.
 * P(N,L) = P[N*P_length_size+L]
 */
static double *P = 0;

/*
 * Dimensions of P (exclusive).
 */
static int P_length_size = 0;
static int P_mismatches_size = 0;

/*
 * Maximum values stored in P (inclusive). 
 */
static int P_max_length = 0;
static int P_max_mismatches = 0;

/*
 * A cluster of N mismatches (including gap blocks) is considered significantly compact,
 * if it's length is not greater than max_sig_cluster_sizes[N].
 */
static int *max_sig_cluster_sizes = 0;

/*
 * A variant (mismatch or gap block).
 */
typedef struct VAR
{
    long long start;  /* Start coordinate (0-based) in the alignment. */
    long long end;    /* End coordinate (0-based) in the alignment. */
    long long len;    /* Variant lenght. */
} VAR;

static VAR *vars = 0;            /* Array of variants for current alignment. */
static long long vars_size = 0;  /* Size of currently allocated 'vars', in variants. */
static long long n_vars = 0;     /* Number of variants stored currently. */



/*
 * Statistics.
 */

/*
 * Number of alignments, processed and skipped (total, i.e., for all processed files).
 */
static long long n_alignments_scanned_total = 0ll;
static long long n_alignments_skipped_total = 0ll;
static long long n_alignments_looked_at_total = 0ll;

/*
 * Total length of all processed alignments.
 */
static long long scanned_alignment_length_total = 0ll;

/*
 * Counting mismatches and gaps in all scanned alignments.
 */
static long long n_vars_total = 0ll;
static long long n_substitutions_total = 0ll;
static long long n_gaps_total = 0ll;
static long long n_gap_blocks_total = 0ll;

/*
 * Total number of mismatch-rich regions found in all scanned alignments.
 */
static long long n_regions_total = 0ll;




/*
 * Stores a complimentary nucleotide for each nucleotide.
 */
static int complimentary_nucleotide[256];



#endif
