/*
 * interface.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Alignment Mismatch Finder source code,
 * and may only be used, modified, and distributed under the terms of
 * the license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C

void help()
{
    fputs("Use -h for help\n",stderr);
}

void print_version()
{
    static char *version_str =
        PROGRAM_NAME ", version " VERSION ", " RELEASE_DATE "\n"
        "Copyright (c) " COPYRIGHT_YEARS " " COPYRIGHT_AUTHORS "\n";
    fputs(version_str,stderr);
}

void print_help()
{
    static char *help_str =
        "Usage: " EXECUTABLE_NAME " [<options>] [<file>...]\n"
        "Options:\n"
        "  -i, --in-format FORMAT  - Set input format to one of: fasta, nss, axt, g-compass.\n"
        "  -o, --out-format FORMAT - Set output format to one of: fasta, nss, cmeh.\n"
        "  -d, --distance D        - Assume that distance between sequences is D.\n"
        "  -e, --evalue E          - Use E as expect value for variant-rich regions (default: 0.01).\n"
        "  -L, --min-aln-length N  - Only scan alignments that are at least N bp long (default: 1).\n"
        "  -n, --min-vars N        - Scan for clusters of at least N variants (substitutions and gap blocks) (default: 3).\n"
        "  -m, --max-vars M        - Scan for clusters of at most M variants (default: same with --min-vars).\n"
        "  -x, --extend N          - Extend each region by N bases on both ends before printing (default: 0).\n"
        "  -j, --join              - Extended regions may overlaps, with '--join' such regions are joined together.\n"
        "  -t, --trim N            - Ignore mismatches/gaps within first and last N bp of each alignment (default: 0).\n"
        "  -g, --use-gaps          - Use gaps together with mismatches (by default only mismatches are used).\n"
        "  -c, --ignore-CpG        - Ignore mismatches/gaps at CpG sites.\n"
        "  -p, --palindrome        - Search for palindromes surrounded by gaps.\n"
        "  -v, --version           - Show version.\n"
        "  -h, --help              - Show help.\n";
    fputs(help_str,stderr);
}

enum al_format parse_alignment_format_name(char *name)
{
    enum al_format format;
    assert(name);
    assert(*name);
    for (format=al_format_unspecified+1; format<n_al_formats && strcasecmp(name,al_format_names[format]); format++) {}
    return format;
}

/*
 * Converts a zero-terminated string to an unsigned 31-bit integer,
 * taking care of any possible misformatting or overflow.
 * Returns: The number, or 0 on failure.
 */
unsigned parse_uint31(char *str,char *name,int talk,int quit)
{
    char *start = str;
    char *c;
    unsigned total = 0;
    static const char cant_parse[] = "Can't parse %s (\"%s\")\n";

    assert(str);
    assert(name);

    while (isspace(*(unsigned char*)start)) { start++; }
    if (*start == '+') { start++; }
    while (*start == '0') { start++; }

    if (!*start)
    {
        if (start != str && *(start-1) == '0') { return 0; }
        if (talk) { fprintf(stderr,cant_parse,name,str); }
        if (quit) { exit(1); } else { return 0; }
    }

    for (c = start; *c; c++)
    {
        if (!isdigit(*(unsigned char*)c))
        {
            if (talk) { fprintf(stderr,cant_parse,name,str); }
            if (quit) { exit(1); } else { return 0; }
        }
    }

    if (c <= start)
    {
        if (talk) { fprintf(stderr,cant_parse,name,str); }
        if (quit) { exit(1); } else { return 0; }
    }

    if (strlen(start) > 10 || (strlen(start) == 10 && strcmp(start,"2147483647") > 0))
    {
        if (talk) { fprintf(stderr,cant_parse,name,str); }
        if (quit) { exit(1); } else { return 0; }
    }

    for (c = start; *c; c++)
    {
        total = total * 10 + (*c - '0');
    }

    return total;
}

double parse_double(char *str,char *name)
{
    char *endptr;
    double a;
    assert(str);
    assert(*str);
    assert(name);
    a = strtod(str,&endptr);
    if ((a == HUGE_VAL || a == -HUGE_VAL) && errno == ERANGE) { fprintf(stderr,"Error: %s is out of range (\"%s\")\n",name,str); exit(1); }
    if (*endptr) { fprintf(stderr,"Can't parse the %s (\"%s\")\n",name,str); exit(1); }
    return a;
}

void parse_options(int argc,char **argv)
{
    enum opt_codes { oc_input_format = 256, oc_output_format, oc_distance, oc_min_vars, oc_max_vars, oc_evalue, oc_extend, oc_trim, oc_min_aln_length };
    static struct option long_options[] = {
        {"version"       , no_argument, &opt_version             , 1},
        {"help"          , no_argument, &opt_help                , 1},
        {"use-gaps"      , no_argument, &use_gaps                , 1},
        {"ignore-CpG"    , no_argument, &ignore_CpG              , 1},
        {"palindrome"    , no_argument, &scan_for_palindromes    , 1},
        {"join"          , no_argument, &join_overlapping_regions, 1},
        {"in-format"     , required_argument, 0, oc_input_format  },
        {"out-format"    , required_argument, 0, oc_output_format },
        {"distance"      , required_argument, 0, oc_distance      },
        {"min-vars"      , required_argument, 0, oc_min_vars      },
        {"max-vars"      , required_argument, 0, oc_max_vars      },
        {"evalue"        , required_argument, 0, oc_evalue        },
        {"extend"        , required_argument, 0, oc_extend        },
        {"trim"          , required_argument, 0, oc_trim          },
        {"min-aln-length", required_argument, 0, oc_min_aln_length}, 
        {0, 0, 0, 0} };
    int c;
    opterr = 0;
    do {
        int curind = optind;
        c = getopt_long(argc,argv,"vhcgji:o:d:n:m:e:x:t:L:",long_options,0);
        switch (c)
        {
            case 'v': opt_version = 1; break;
            case 'h': opt_help = 1; break;
            case 'g': use_gaps = 1; break;
            case 'c': ignore_CpG = 1; break;
            case 'p': scan_for_palindromes = 1; break;
            case 'j': join_overlapping_regions = 1; break;
            case 'i': case oc_input_format: input_format = parse_alignment_format_name(input_format_str = optarg); break;
            case 'o': case oc_output_format: output_format = parse_alignment_format_name(output_format_str = optarg); break;
            case 'd': case oc_distance: dist = parse_double(dist_str = optarg,"distance"); break;
            case 'n': case oc_min_vars: min_vars_per_cluster = parse_uint31(optarg,"min vars",1,1); break;
            case 'm': case oc_max_vars: max_vars_per_cluster = parse_uint31(optarg,"max vars",1,1); break;
            case 'e': case oc_evalue: evalue_str = optarg; evalue = parse_double(optarg,"evalue"); break;
            case 'x': case oc_extend: extend_output_regions = parse_uint31(optarg,"extend",1,1); break;
            case 't': case oc_trim: trim_alignment = parse_uint31(optarg,"trim",1,1); break;
            case 'L': case oc_min_aln_length: min_alignment_length = parse_uint31(optarg,"minimum alignment length",1,1); break;
            case '?':
                if (optopt) { fprintf(stderr,"Unrecognized option -%c\n",optopt); }
                else { fprintf(stderr,"Unrecognized option \"%s\"\n",argv[curind]); }
                help();
                exit(1);
                break;
            case 0: break;
            case -1: break;
            default: assert(0);
        }
    }
    while (c != -1);

    if (optind > 0 && optind < argc) { argv_index_of_first_path = optind; }
}

void read_command_line(int argc,char **argv)
{
    parse_options(argc,argv);
    if (opt_version) { print_version(); }
    if (opt_help) { print_help(); }

    if (input_format == al_format_unspecified)
    {
        if (!opt_version && !opt_help) { fprintf(stderr,"Input format is not specified (\"--input-format\" option)\n"); help(); }
        exit(1);
    }
    if (input_format == al_format_unknown)
    {
        fprintf(stderr,"Error: Unknown input format \"%s\"\n",input_format_str);
        exit(1);
    }
    if (input_format != al_format_axt &&
        input_format != al_format_fasta &&
        input_format != al_format_g_compass &&
        input_format != al_format_nss)
    {
        fprintf(stderr,"Error: Loading alignment in %s format is not supported\n",al_format_names[input_format]);
        exit(1);
    }

    if (output_format == al_format_unspecified)
    {
        if (!opt_version && !opt_help) { fprintf(stderr,"Output format is not specified (\"--output-format\" option)\n"); help(); }
        exit(1);
    }
    if (output_format == al_format_unknown)
    {
        fprintf(stderr,"Error: Unknown output format \"%s\"\n",output_format_str);
        exit(1);
    }
    if (output_format != al_format_fasta &&
        output_format != al_format_nss &&
        output_format != al_format_cmeh)
    {
        fprintf(stderr,"Error: Writing output in %s format is not supported\n",al_format_names[output_format]);
        exit(1);
    }

    if (!dist_str)
    {
        fprintf(stderr,"Error: Assumed evolutionary distance is not specified (\"--distance\" option)\n");
        exit(1);
    }
    if (dist <= 0)
    {
        fprintf(stderr,"Error: Assumed evolutionary distance is not positive (\"%s\")\n",dist_str);
        exit(1);
    }

    if (min_alignment_length < 1ll) { min_alignment_length = 1ll; }

    if (min_vars_per_cluster < 0) { min_vars_per_cluster = DEFAULT_MIN_VARS_PER_CLUSTER; }
    if (min_vars_per_cluster < 2) { min_vars_per_cluster = 2; }
    if (min_vars_per_cluster > 1000) { min_vars_per_cluster = 1000; }
    /*if (max_vars_per_cluster < 0) { max_vars_per_cluster = min_vars_per_cluster; }*/
    if (max_vars_per_cluster < min_vars_per_cluster) { max_vars_per_cluster = min_vars_per_cluster; }
    if (max_vars_per_cluster > 1000) { max_vars_per_cluster = 1000; }
    if (evalue <= 0.0)
    {
        fprintf(stderr,"Error: Evalue is not positive (%s)\n",evalue_str);
        exit(1);
    }
    if (evalue >= 1.0)
    {
        fprintf(stderr,"Error: Evalue should be smaller than 1 (%s)\n",evalue_str);
        exit(1);
    }

    if (!use_gaps && scan_for_palindromes)
    {
        fprintf(stderr,"Warning: --palindrome has no effect without --use-gaps\n");
        scan_for_palindromes = 0;
    }
}

#endif
