/*
 * main.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Alignment Mismatch Finder source code,
 * and may only be used, modified, and distributed under the terms of
 * the license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C

/*
 * Exit handler, deallocates memory.
 */
void done()
{
    if (P) { free1(P,sizeof(*P)*P_length_size*P_mismatches_size); P = 0; }
    if (max_sig_cluster_sizes) { free1(max_sig_cluster_sizes,sizeof(*max_sig_cluster_sizes)*P_mismatches_size); max_sig_cluster_sizes = 0; }
    if (vars) { free1(vars,sizeof(*vars)*vars_size); vars = 0; }
    if (allocated > 0) { fprintf(stderr,"Error: Failed to free %lld bytes\n",allocated); }
}

/*
 * Initializing arrays P and max_sig_cluster_sizes.
 */
void scan_init()
{
    int N,L;

    /*
     * Actually it's not necessary to have the whole P matrix in memory if all we want is to find
     * the length boundary for given d and N.
     * However this matrix helps to determine and report significance of located mismatch-rich regions.
     *
     * To do: instead of allocating and filling the whole matrix in the beginning,
     * re-allocate and fill the matrix as needed.
     */
    P_max_length = 10000;
    P_max_mismatches = 100;
    P_length_size = P_max_length + 1;
    P_mismatches_size = P_max_mismatches + 1;
    P = (double*)malloc1(sizeof(*P)*P_length_size*P_mismatches_size);

    /* Probability of having at 0 or more mismatches is 1 for any length. */
    for (L=0; L<=P_max_length; L++) { P[L] = 1.0; }

    /* Probability of having 1 or more mismatches in 0 bp region is 0. */
    for (N=1; N<=P_max_mismatches; N++) { P[N*P_max_length] = 0.0; }

    /* Filling the matrix P using the formula: P(N,L) = P(N,L-1)*(1-d) + P(N-1,L-1)*d. */
    for (N=1; N<=P_max_mismatches; N++)
    {
        for (L=1; L<=P_max_length; L++)
        {
            P[N*P_length_size+L] = P[N*P_length_size+L-1]*(1-dist) + P[(N-1)*P_length_size+L-1]*dist;
        }
    }

    /* Printing a fragment of matrix P for checking. */
    /*printf("      ");
    for (N=0; N<=3; N++) { printf(" %7d",N); }
    printf("\n");
    for (L=0; L<=40; L++)
    {
        printf("%5d ",L);
        for (N=0; N<=3; N++) { printf(" %7.5f",P[N*P_length_size+L]); }
        printf("\n");
    }*/

    /* Computing the largest size of significant cluster of N mismatches, for each N. */
    max_sig_cluster_sizes = malloc1(sizeof(*max_sig_cluster_sizes)*P_mismatches_size);
    max_sig_cluster_sizes[0] = 0;
    for (N=1; N<=P_max_mismatches; N++)
    {
        for (L=0; L <= P_max_length && P[N*P_length_size+L] <= evalue; L++) {}
        max_sig_cluster_sizes[N] = L-1;
    }

    /* Printing max_sig_cluster_sizes. */
    /*printf("A cluster of mismatches is significantly rare if:\n");
    for (N=2; N<=10; N++)
    {
        if (N > max_sig_cluster_sizes[N]) {}
        else if (N == max_sig_cluster_sizes[N]) { printf("    - if it's a continuous stretch of %d mismatches\n",N); }
        else { printf("    - it has %d mismatches spread over at most %d bp\n",N,max_sig_cluster_sizes[N]); }
    }*/

    /* Setting up the table of complimentary nucleotides. */
    {
        int i;
        for (i=0; i<256; i++) { complimentary_nucleotide[i] = i; }
        complimentary_nucleotide['a'] = 't';
        complimentary_nucleotide['t'] = 'a';
        complimentary_nucleotide['c'] = 'g';
        complimentary_nucleotide['g'] = 'c';
        complimentary_nucleotide['A'] = 'T';
        complimentary_nucleotide['T'] = 'A';
        complimentary_nucleotide['C'] = 'G';
        complimentary_nucleotide['G'] = 'C';
    }
}

int main(int argc,char **argv)
{
    atexit(done);

#ifdef DEBUG
    setbuf(stdout,0);
    setbuf(stderr,0);
#endif

    read_command_line(argc,argv);

    fprintf(stderr,"Looking for clusters of %d",min_vars_per_cluster);
    if (max_vars_per_cluster > min_vars_per_cluster) { fprintf(stderr,"-%d",max_vars_per_cluster); }
    fprintf(stderr," %s, expected to appear with probability %.5f (assuming distance %.5f)\n",use_gaps?"mismatches or gap blocks":"mismatches",evalue,dist);
    if (min_alignment_length > 1) { fprintf(stderr,"Scanning only alignments that are at least %lld bp long\n",min_alignment_length); }

    scan_init();

    if (argv_index_of_first_path > 0) { int i; for (i=argv_index_of_first_path; i<argc; i++) { scan_file(argv[i]); } }
    else { scan_file(0); }

    fprintf(stderr,"Scanned %lld alignments (total length: %lld bp)",n_alignments_scanned_total,scanned_alignment_length_total);
    if (n_alignments_skipped_total) { fprintf(stderr,", skipped %lld alignments",n_alignments_skipped_total); }
    fputs("\n",stderr);

    fprintf(stderr,"Observed distance between sequences");
    if (n_alignments_scanned_total > 1) { fprintf(stderr," (based on all alignments)"); }
    fprintf(stderr,": %.5f variants / site\n",((double)(n_substitutions_total+n_gap_blocks_total))/(scanned_alignment_length_total-n_gaps_total/2));
    fprintf(stderr," => In total %lld variants were found (%lld substitutions, %lld gap blocks)\n",n_vars_total,n_substitutions_total,n_gap_blocks_total);
    fprintf(stderr," => %lld regions were found\n",n_regions_total);

    return 0;
}

#endif
