/*
 * scan.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Alignment Mismatch Finder source code,
 * and may only be used, modified, and distributed under the terms of
 * the license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C

/*
 * Checks whether sequence between 'start' and 'end' (both inclusive) is a text palindrome.
 * (Text palindrome is an ordinary palindrome - a text that reads the same way in both directions.)
 * Returns 1 if palindrome, 0 otherwise.
 */
int string_is_a_text_palindrome(char *start,char *end)
{
    char *c = start, *d = end;

    for (; c < d; c++,d--)
    {
        int uc = toupper(*(unsigned char*)c);
        int ud = toupper(*(unsigned char*)d);

        if (uc == 'N' || uc == 'X' || ud == 'N' || ud == 'X') { continue; }
        if (uc != ud) { return 0; }
    }
    return 1;
}

/*
 * Checks whether sequence between 'start' and 'end' (both inclusive) is a genomic palindrome.
 * (Genomic palindrome is sequence that reads the same on both strands, for example 'CCGG')
 * Returns 1 if palindrome, 0 otherwise.
 */
int string_is_a_genomic_palindrome(char *start,char *end)
{
    char *c = start, *d = end;

    for (; c < d; c++,d--)
    {
        int uc = toupper(*(unsigned char*)c);
        int ud = toupper(complimentary_nucleotide[*(unsigned char*)d]);

        if (uc == 'N' || uc == 'X' || ud == 'N' || ud == 'X') { continue; }
        if (uc != ud) { return 0; }
    }
    return 1;
}

/*
 * Scans given alignment for mismatch-rich regions.
 */
void scan_alignment(taln *aln)
{
    long long len,pos;
    int i;
    int region_first_var = -1;
    int region_last_var = -1;
    int region_end = -1;
    int prev_gap_seq = 2;  /* 0 = gap in first sequence, 1 = gap in second sequence, 2 = no gap. */

    assert(aln);

    len = aln->seq[0].data.length;
    if (len <= 0) { return; }

    /*{
        printf("\nScanning alignment\n");
        int i;
        for (i=0; i<2; i++)
        {
            printf("    \"");
            fwrite(aln->seq[i].name.start,1,aln->seq[i].name.length,stdout);
            printf("\": %" PRINT_SIZE_T " bp (%" PRINT_SIZE_T " bytes)\n",aln->seq[i].data.length,aln->seq[i].data.size);
        }
    }*/

    /* Extending array of variants if new alignment is longer than previous ones. */
    if (vars_size < len)
    {
        if (vars) { free1(vars,sizeof(*vars)*vars_size); }
        while (vars_size < len) { vars_size += 1000; }
        vars = malloc1(sizeof(*vars)*vars_size);
    }

    /* Making a list of all variants in this alignment. */
    n_vars = 0;
    for (pos=trim_alignment; pos<len-trim_alignment*2; pos++)
    {
        int c1 = toupper(((unsigned char*)aln->seq[0].data.start)[pos]);
        int c2 = toupper(((unsigned char*)aln->seq[1].data.start)[pos]);
        int gap_seq = (c1 == '-') ? 0 : (c2 == '-') ? 1 : 2;

        /* Skipping CpG site. */
        if (ignore_CpG && pos < (long long)aln->seq[0].data.length-1)
        {
            if ( (c1 == 'C' && toupper(((unsigned char*)aln->seq[0].data.start)[pos+1]) == 'G') ||
                 (c2 == 'C' && toupper(((unsigned char*)aln->seq[1].data.start)[pos+1]) == 'G') ) { pos++; continue; }
        }

        if (gap_seq < 2)  /* Gap is found! */
        {
            if (use_gaps)
            {
                if (gap_seq == prev_gap_seq && n_vars > 0)
                {
                    /* Gap on same sequence => Extending gap. */
                    assert(n_vars-1 >= 0);
                    assert(n_vars-1 < vars_size);
                    assert(vars[n_vars-1].end == pos-1);
                    vars[n_vars-1].end = pos;
                    vars[n_vars-1].len++;
                }
                else
                {
                    /* New gap or gap on opposite sequence => Opening new gap. */
                    assert(n_vars < vars_size);
                    vars[n_vars].start = pos;
                    vars[n_vars].end = pos;
                    vars[n_vars].len = 1;
                    n_vars++;
                    n_gap_blocks_total++;
                }
                n_gaps_total++;
            }
        }
        else   /* No gap. */
        {
            if (c1 != c2 && c1 != 'X' && c1 != 'N' && c2 != 'X' && c2 != 'N')
            {
                assert(n_vars < vars_size);
                vars[n_vars].start = pos;
                vars[n_vars].end = pos;
                vars[n_vars].len = 1;
                n_vars++;
                n_substitutions_total++;
            }
        }

        prev_gap_seq = gap_seq;
    }
    n_vars_total += n_vars;

    /*
     * Detecting significantly unusual clusters of variants.
     * Merging overlapping clusters into regions.
     */
    for (i=0; i<=n_vars-2; i++)
    {
        int cluster_last_var = i;
        int cluster_end;
        int n;

        assert(i < vars_size);
        cluster_end = vars[i].end;

        /*
         * Checking if cluster of n mismatches is statistically significant,
         * for each n from --min-vars to --max-vars.
         * (Plus an optional special case of 2 gaps enclosing a palindrome).
         */
        for (n=min_vars_per_cluster; (n <= max_vars_per_cluster) && (i+n <= n_vars); n++)
        {
            if (vars[i+n-1].start - vars[i].end + 1 <= max_sig_cluster_sizes[n])
            {
                cluster_last_var = i+n-1;
                cluster_end = vars[cluster_last_var].end;
            }
        }

        /* Checking for palindrome. */
        if ( cluster_last_var == i && scan_for_palindromes &&
             ( (aln->seq[0].data.start[vars[i].end] == '-' && aln->seq[1].data.start[vars[i+1].start] == '-') ||
               (aln->seq[1].data.start[vars[i].end] == '-' && aln->seq[0].data.start[vars[i+1].start] == '-') ) )
        {
            int s = vars[i].end + 1;
            int e = vars[i+1].start - 1;
            if (e-s < 2 ||
                string_is_a_text_palindrome   (aln->seq[0].data.start+s,aln->seq[0].data.start+e) ||
                string_is_a_genomic_palindrome(aln->seq[0].data.start+s,aln->seq[0].data.start+e) ) { cluster_last_var = i+1; }
        }

        /*
         * If any new cluster is found, checking if it can be merged with previous region,
         * or if it starts a new region.
         */
        if (cluster_last_var > i)
        {
            if (region_end + (join_overlapping_regions ? extend_output_regions*2 : 0) < vars[i].start)
            {
                write_alignment_region(aln,region_first_var,region_last_var);
                region_first_var = i;
            }
            region_last_var = cluster_last_var;
            region_end = cluster_end;
        }
    }
    write_alignment_region(aln,region_first_var,region_last_var);

    scanned_alignment_length_total += len;

    /*printf("Scanning complete\n");*/
}

/*
 * Scans the alignment, reading it from 'file'.
 */
void scan_buffered_file(tfile *file)
{
    taln aln;
    long long n_alignments_scanned_this_file = 0ll;
    long long n_alignments_skipped_this_file = 0ll;
    long long file_sequence_length = 0ll;
    long long file_names_length = 0ll;
    long long file_info_length = 0ll;

    assert(file);

    init_aln(&aln);
    while (load_alignment(&aln,file))
    {
        file_info_length += aln.info.length;
        file_sequence_length += aln.seq[0].data.length;
        file_sequence_length += aln.seq[1].data.length;
        file_names_length += aln.seq[0].name.length;
        file_names_length += aln.seq[1].name.length;

        if (!aln.seq[0].data.length || !aln.seq[1].data.length ||
            aln.seq[0].data.length != aln.seq[1].data.length ||
            (long long)aln.seq[0].data.length < min_alignment_length)
        {
            n_alignments_looked_at_total++;
            n_alignments_skipped_this_file++;
            n_alignments_skipped_total++;
            continue;
        }

        scan_alignment(&aln);

        n_alignments_looked_at_total++;
        n_alignments_scanned_this_file++;
        n_alignments_scanned_total++;
    }
    clean_aln(&aln);

    fprintf(stderr," .. %lld bp of sequence, %lld bytes of names, %lld bytes of info\n",
            file_sequence_length,file_names_length,file_info_length);
    if (n_alignments_skipped_this_file)
    {
        fprintf(stderr,"    Skipped %lld out of %lld alignments\n",
                n_alignments_skipped_this_file,n_alignments_skipped_this_file+n_alignments_scanned_this_file);
    }
}

/*
 * Scans the alignment file specified in 'path'.
 * If 'path' is 0, processes standard input.
 */
void scan_file(char* path)
{
    tfile file = {0,0,0,0,0};

    if (path)
    {
        tfile *f = open_file(&file,path,"rb",BUFSIZE);
        if (!f) { fprintf(stderr,"Error: Can't open \"%s\"\n",path); return; }
        fprintf(stderr,"Scanning \"%s\"\n",path);
    }
    else
    {
        make_buffered_file_from_stream(&file,stdin,BUFSIZE);
        fprintf(stderr,"Scanning standard input\n");
    }

    scan_buffered_file(&file);

    close_file(&file);
    free_file_buffer(&file);
}

#endif
